#include <sched.h>
#include "PhysicalDIO.h"
#include "FPGA.h"

static DIOAPI PhysicalDIOAPI = {
  .Init = (void *)PhysicalDIOInit,
  .Fini = (void *)PhysicalDIOFini,
  .Lock = (void *)PhysicalDIOLock,
  .Unlock = (void *)PhysicalDIOUnlock,
  .Preempt = (void *)PhysicalDIOPreempt,
  .Refresh = (void *)PhysicalDIORefresh,
  .Commit = (void *)PhysicalDIOCommit,
  .Set = (void *)PhysicalDIOSet,
  .Get = (void *)PhysicalDIOGet,
  .SetAsync = (void *)PhysicalDIOSetAsync,
  .GetAsync = (void *)PhysicalDIOGetAsync,
  .Count = (void *)PhysicalDIOCount,
  .Capabilities = (void *)PhysicalDIOCapabilities
};

void *PhysicalDIOInit(PhysicalDIO *dio,void *bus1,void *sub1) {
  Bus *bus = bus1;
  DIORaw *sub = sub1;
  int err;

  if (dio->InitStatus > 0) return dio;
  dio->API = &PhysicalDIOAPI;
  dio->bus = bus;
  dio->dio = sub;
  dio->NumLocks = sub->API->Count(sub);
  if (dio->bus->InitStatus <= 0
      ||dio->dio->InitStatus <= 0) {
    dio->API->Fini(dio);
    dio->InitStatus = -1;
    return dio;
  }
  if ((err=FPGAReloadRegister(&dio->gotHUP,7)) < 0) {
    dio->API->Fini(dio->API);
    dio->InitStatus = err;
    return dio;
  }
  dio->InitStatus = 1;
  return dio;
}

void PhysicalDIOFini(PhysicalDIO *dio) {
  if (dio->InitStatus > 0) {
    FPGAReloadUnregister(&dio->gotHUP);
  }
  dio->bus->API->Fini(dio->bus);
  if (dio->InitStatus > 0) dio->InitStatus = 0;
}

void PhysicalDIORefresh(PhysicalDIO *dio) {
  dio->bus->API->Lock(dio->bus,0,0);
  if (dio->gotHUP) {
    dio->bus->API->Commit(dio->bus,1);
    dio->gotHUP = 0;
  }
  dio->bus->API->Refresh(dio->bus);
  dio->bus->API->Unlock(dio->bus,0,0);
}

void PhysicalDIOCommit(PhysicalDIO *dio,int forceall) {
  if (dio->gotHUP) {
    forceall = 1;
    dio->gotHUP = 0;
  }
  dio->bus->API->Lock(dio->bus,0,0);
  dio->bus->API->Commit(dio->bus,forceall);
  dio->bus->API->Unlock(dio->bus,0,0);
}

static inline void _DIOSet(DIORaw *dio,int Num,DIOState State) {
  if (State < 0) {
    dio->API->DirSet(dio,Num,0);
  } else {
    // Set data first to avoid glitch if current state doesn't match new state
    dio->API->DataSet(dio,Num,State==HIGH);
    dio->API->DirSet(dio,Num,1);
  }
}

static inline DIOState _DIOGet(DIORaw *dio,int Num) {
  if (dio->API->DirGet(dio,Num)) {
    return dio->API->DataGet(dio,Num);
  } else {
    return INPUT_LOW + dio->API->DataGet(dio,Num);    
  }
}

void PhysicalDIOSet(PhysicalDIO *dio,int DIONum,DIOState State) {
  if (DIONum < 0 || DIONum > dio->NumLocks 
      || dio->Capabilities[DIONum] == DIO_NOT_PRESENT) return;
  dio->bus->API->Lock(dio->bus,0,0);
  _DIOSet(dio->dio,~DIONum,State);
  dio->bus->API->Unlock(dio->bus,0,0);
}

DIOState PhysicalDIOGet(PhysicalDIO *dio,int DIONum) {
  DIOState ret;
  if (DIONum < 0 || DIONum > dio->NumLocks 
      || dio->Capabilities[DIONum] == DIO_NOT_PRESENT) return INPUT;
  dio->bus->API->Lock(dio->bus,0,SHARED);
  ret =  _DIOGet(dio->dio,~DIONum);
  dio->bus->API->Unlock(dio->bus,0,SHARED);
  return ret;
}

void PhysicalDIOSetAsync(PhysicalDIO *dio,int DIONum,DIOState State) {
  if (DIONum < 0 || DIONum > dio->NumLocks 
      || dio->Capabilities[DIONum] == DIO_NOT_PRESENT) return;
  dio->bus->API->Lock(dio->bus,0,0);
  _DIOSet(dio->dio,DIONum,State);
  dio->bus->API->Unlock(dio->bus,0,0);
}

DIOState PhysicalDIOGetAsync(PhysicalDIO *dio,int DIONum) {
  DIOState ret;

  //printf("%d:%d:%d\n",DIONum,dio->NumLocks,dio->Capabilities[DIONum] == DIO_NOT_PRESENT);
  if (DIONum < 0 || DIONum > dio->NumLocks 
      || dio->Capabilities[DIONum] == DIO_NOT_PRESENT) return INPUT;
  dio->bus->API->Lock(dio->bus,0,SHARED);
  ret = _DIOGet(dio->dio,DIONum);
  dio->bus->API->Unlock(dio->bus,0,SHARED);
  return ret;
}

int PhysicalDIOLock(PhysicalDIO *dio,unsigned num,int flags) {
  return 1;
}
int PhysicalDIOUnlock(PhysicalDIO *dio,unsigned num,int flags) {
  return 1;
}
void PhysicalDIOPreempt(PhysicalDIO *dio) {
}

int PhysicalDIOWait(PhysicalDIO *dio,int *match,int min,int max,int nh,
		    int *h,int nl,int *l) {
  int i,matched,state;

  do {
    matched = 0;
    dio->bus->API->Lock(dio->bus,0,SHARED);
    PhysicalDIORefresh(dio);
    for (i=0;i<nh;i++) {
      state = _DIOGet(dio->dio,~h[i]);
      if (state == HIGH || state == INPUT_HIGH) {
	match[matched++] = h[i];
      }
    }
    for (i=0;i<nl;i++) {
      state = _DIOGet(dio->dio,~l[i]);
      if (state == LOW || state == INPUT_LOW) {
	match[matched++] = l[i];
      }
    }
    dio->bus->API->Unlock(dio->bus,0,SHARED);
    if (matched < min || matched > max) sched_yield();
  } while (matched < min || matched > max);
  return matched;
}

unsigned PhysicalDIOCount(PhysicalDIO *dio) {
  return dio->dio->API->Count(dio);
}

DIOCaps PhysicalDIOCapabilities(PhysicalDIO *dio,unsigned num) {
  if (num < 0 || num >= dio->NumLocks) return DIO_NOT_PRESENT;
  return dio->Capabilities[num];
}

// Author: Michael Schmidt (michael@embeddedARM.com)
// Copyright (c) 2011, Technologic Systems, All Rights Reserved
// Refer to the COPYRIGHT file provided with this project for licensing terms.
